""" http://www.sphinx-doc.org/en/stable/extdev/markupapi.html
"""
import json
from functools import partial
from io import StringIO
from typing import List, Union, Dict, Sequence

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged
from docutils.statemachine import ViewList
from drf_openapi.entities import OpenApiSchemaGenerator
from sphinx.util.docstrings import prepare_docstring
from sphinx.util.nodes import nested_parse_with_titles

from frameapp.schemas import json_schema_str

from frameapp.configurator import Configurator


class APIDocDirective(Directive):
    required_arguments = 0
    optional_arguments = 0
    has_content = False  # Whether a directive source markup has a body with some content

    option_spec = {
        'prefix': unchanged
    }

    def run(self) -> List:
        """ http://www.sphinx-doc.org/en/stable/extdev/markupapi.html#docutils.parsers.rst.Directive.run
        """
        env = self.state.document.settings.env
        app = env.app
        config: Configurator = app.config.FRAMEAPP_CONFIG
        prefix = self.options.get('prefix', '/')

        app.info('Generating API documentation from gathered project metadata')

        doc_text = StringIO()
        api_doc = []
        doc_append = api_doc.append

        for namespace, route in config.routes.registry.items():
            # doc_text.write(namespace_caption(f'Namespace: {namespace}'))
            # doc_text.write(line_separator())

            for route_name, route_struct in route.items():
                if not route_struct.pattern.startswith(prefix):
                    continue

                pattern = route_struct.pattern[len(prefix):]

                doc_text.write(url_pattern_caption(pattern))
                doc_text.write(line_separator())

                for schema_sig, schema_data in route_struct.schemas.items():
                    doc_text.write(api_version(schema_sig.api_version))
                    doc_text.write(line_separator())

                    doc_text.write(request_method(schema_sig.request_method))
                    doc_text.write(line_separator())

                    docstring = '\n'.join(prepare_docstring(schema_data.doc))
                    doc_text.write(docstring)
                    doc_text.write(line_separator())

                    if schema_data.input_schema:
                        doc_text.write(output_schema(schema_data.output_schema))
                        doc_text.write(line_separator())

                    if schema_data.input_serializer:
                        # Let's create a temp schema generator. We don't have to care about initial arguments too much
                        version = '1.10'
                        temp = OpenApiSchemaGenerator(version, '<Title>', '/url')
                        response_schema, error_status_codes = temp.get_response_object(schema_data.input_serializer.get(version), '<empty doc>')
                        doc_text.write(input_serializer(response_schema))
                        doc_text.write(line_separator())

                    if schema_data.output_serializer:
                        # Let's create a temp schema generator. We don't have to care about initial arguments too much
                        version = '1.10'
                        temp = OpenApiSchemaGenerator(version, '<Title>', '/url')
                        response_schema, error_status_codes = temp.get_response_object(schema_data.output_serializer.get(version), '<empty doc>')
                        doc_text.write(output_serializer(response_schema))
                        doc_text.write(line_separator())

                    if schema_data.output_schema:
                        doc_text.write(output_schema(schema_data.output_schema))
                        doc_text.write(line_separator())

                    doc_text.write(horizontal_separator())
                    doc_text.write(line_separator())

        # Since there's no extensive guides on how to use nested_parse(),
        # the current implementation just repeats the steps from standard extensions, such as
        # https://github.com/sphinx-doc/sphinx/blob/228fdb892af25f4b93f2760f2cd6497f2aabc0be/sphinx/ext/autosummary/__init__.py#L377-L384
        # See also http://www.sphinx-doc.org/en/stable/extdev/markupapi.html#viewlists
        doc_text.seek(0)
        vl = ViewList(doc_text.readlines())
        output = nodes.paragraph()
        nested_parse_with_titles(self.state, vl, output)
        doc_append(output)

        return api_doc


def line_separator() -> str:
    return '\n\n\n'


def namespace_caption(name: str) -> str:
    hdr = '=' * len(name)
    return f'{hdr}\n{name}\n{hdr}'


def url_pattern_caption(pattern: str) -> str:
    hdr = '-' * len(pattern)
    return f'{pattern}\n{hdr}'


def schema_block(caption: str, schema: Union[str, Dict]) -> str:
    if isinstance(schema, dict):
        schema = json.dumps(schema, indent=2)
    else:
        if schema.endswith('.json'):  # schema is a file path
            schema = json_schema_str(schema)

    buf = StringIO(schema)
    offset_aligned_schema = [f'    {line}' for line in buf.readlines()]
    schema = ''.join(offset_aligned_schema)

    return f'**{caption}:**\n\n.. code-block:: json\n    \n{schema}'


input_schema = partial(schema_block, 'Input JSON Schema')
output_schema = partial(schema_block, 'Output JSON Schema')

input_serializer = partial(schema_block, 'Input Serializer')
output_serializer = partial(schema_block, 'Output Serializer')


def api_version(value: str) -> str:
    return f'**API Version:** {value}'


def request_method(value: Sequence[str]) -> str:
    val = ', '.join(value)
    return f'**Allowed Request Methods:** {val}'


def horizontal_separator():
    return '.. raw:: html\n   \n   <hr>'



from collections import OrderedDict
from typing import NamedTuple, Dict

from .configurator.routes import Route
from .configurator.sums import SumTypeMetaData
from .structures import CompactSerializable


class AppRegistry(NamedTuple):
    """ Application Registry expressed as an immutable structure for use after configuration step.
    """
    routes: Dict[str, Dict[str, Route]]
    sums: Dict[str, SumTypeMetaData]


class _RegistryBuilder(CompactSerializable):
    """ Registry Builder is only used during configuration step, where mutability is convenient.
    """
    __slots__ = ('routes', 'sums')

    def __init__(self) -> None:
        self.routes: Dict[str, Dict[str, Route]] = OrderedDict()
        self.sums: Dict[str, SumTypeMetaData] = OrderedDict()


class predvalseq(tuple):
    """ This class is a copy of ``pyramid.registry.predvalseq``

    A subtype of tuple used to represent a sequence of predicate values """
    pass



import logging
from typing import Callable

import venusian

from ..configurator import Scanner
from ..exceptions import ConfigurationError


log = logging.getLogger(__name__)


class http_endpoint:
    venusian = venusian

    def __init__(self, **settings):
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def callback(scanner: Scanner, name: str, obj: Callable) -> None:
            view_item = scanner.configurator.views.add_view(view=obj, **settings)
            scanner.configurator.routes.add_route_schema(view=obj, **settings)
            namespace = scanner.configurator.routes.namespace
            try:
                routes_namespace = scanner.configurator.routes.registry[namespace]
            except KeyError:
                raise ConfigurationError(f"Namespace was not included: {namespace}")
            try:
                route = routes_namespace[view_item.route_name]
            except KeyError:
                raise ConfigurationError(
                    f'No route named "{view_item.route_name}" found for view registration within "{namespace}" namespace.'
                )
            try:
                renderer = scanner.configurator.renderers.get_renderer(view_item.renderer)
            except Exception as e:
                raise ConfigurationError(f'An error happened while marking http_endpoint({settings}): {e}')

            view_item = view_item._replace(renderer=renderer)
            route.view_metas.append(view_item)

        info = self.venusian.attach(wrapped, callback, category='frameapp_namespace', depth=depth + 1)
        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__
        return wrapped


class http_defaults(http_endpoint):
    """ This object is a copy of ``pyramid.view.view_defaults``.

    A class :term:`decorator` which, when applied to a class, will
    provide defaults for all view configurations that use the class. This
    decorator accepts all the arguments accepted by
    :meth:`pyramid.view.view_config`, and each has the same meaning.

    See :ref:`view_defaults` for more information.
    """
    def __call__(self, wrapped):
        wrapped.__view_defaults__ = self.__dict__.copy()
        return wrapped



import re
from typing import Dict, List, Optional

from .configurator import Configurator


def setup(package: str, apps: Dict[str, List[Dict[str, str]]], namespace='frameapp', ignore: Optional[List[str]] = None) -> Configurator:
    configurator = Configurator()

    for app_name, entry_points in apps.items():
        for entry_point in entry_points:
            configurator.include(app_name, entry_point['entry_point'], entry_point['url_prefix'], namespace)

    if ignore is None:
        ignore = [
            '.__pycache__',
            # do not scan migrations
            re.compile('^.*\.?migrations\.?.*$').match,
            # do not scan test modules included into the project tree
            re.compile('^.*\.?tests\.?.*$').match,
        ]

    configurator.scan(
        package=package,
        namespace=namespace,
        ignore=ignore
    )
    registry = configurator.freeze()
    return configurator



""" This module is initially taken from Solo

https://github.com/avanov/solo/blob/bf44c527dbe48256d2bd3da463eceeb78d05a38d/solo/configurator/config/sums.py
"""
import logging
import re
from collections import OrderedDict
from typing import Dict, Any, get_type_hints, Type, Iterator, Set, NamedTuple

import venusian

from ..exceptions import ConfigurationError


log = logging.getLogger(__name__)


SUM_TYPE_VARIANT_NAME_RE = re.compile('^[A-Z][0-9A-Z_]*$')


class ProxyContract:
    def __init__(self, parent: 'SumVariant') -> None:
        # We reference the parent and not the parent's contract attribute, because
        # this attribute may be replaced during app initialisation (NamedTuple._replace)
        self.parent = parent

    def __getattr__(self, item):
        return getattr(self.parent.contract, item)


class SumVariant:
    venusian = venusian

    def __init__(self,
                 variant_of: Type['SumType'],
                 name: str,
                 constructor,
                 value,
                 contract: NamedTuple) -> None:
        """
        :param variant_of: class of the SumType this variant instance belongs to
        :param name: name of the variant, as it is defined in the source code (uppercase)
        :param value: variant value
        :param constructor: constructor for a data that this variant can hold
        :param contract: bound terms, will be empty most of the time, and present in constructor only for
        replicating them when SumVariantInstance is created from base SumVariant
        """
        self.variant_of = variant_of
        self.name = name
        self.value = value
        self.constructor = constructor
        self.contract = contract
        # A proxy object to imitate access through SumType.Contract
        # Exists only for the sake of satisfying PyCharm type checker
        self.Contract = ProxyContract(self)

    def is_primitive_type(self) -> bool:
        return isinstance(self.constructor, (int, str, float, bool))

    def __call__(self, *data_args, **data_kwargs) -> 'SumVariantInstance':
        """ Returns a data-holding variant"""
        return SumVariantInstance(data_args,
                                  data_kwargs,
                                  variant_of=self.variant_of,
                                  name=self.name,
                                  constructor=self.constructor,
                                  value=self.value,
                                  contract=self.contract)

    def __eq__(self, other: 'SumVariant') -> bool:
        """ This method is redefined only to simplify variant comparison for tests with mocks that
        might do things like mocked_function.assert_called_with(SumType.VARIANT)
        """
        return all([
            self.variant_of is other.variant_of,
            self.name == other.name,
            self.value == other.value,
            self.constructor == other.constructor,
            self.contract == other.contract
        ])

    # https://docs.python.org/3/reference/datamodel.html#object.__hash__
    # If a class that overrides __eq__() needs to retain the implementation of __hash__() from a parent class,
    # the interpreter must be told this explicitly by setting __hash__ = <ParentClass>.__hash__
    __hash__ = object.__hash__

    def bind(self, contract_term: str) -> Any:
        """ This method is called when we declare variant cases. For instance:
            >>> class Language(SumType):
            >>>     ENGLISH = 'en'
            >>>
            >>> @Language.ENGLISH.bind('term_name')
            >>> def term_implementation():
            >>>     pass
            >>>
        :param contract_term: name of the term
        """
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def wrapper(wrapped):
            def callback(scanner, name, obj):
                """ This method will be called by venusian.
                The ``callback`` name is not reserved, we just pass this callable
                to ``self.venusian.attach()`` below. Although, callback's arguments list is strictly defined.
                """
                attr = settings.get('attr')
                if attr:  # check whether a decorated object is something inside a class scope
                    obj = getattr(obj, attr)

                previous_term = getattr(self.contract, contract_term)
                if previous_term is not None:
                    # venusian may scan the same declarations multiple times during the app initialization,
                    # therefore we allow re-assignment of the same case implementations and prohibit
                    # any new implementations
                    if previous_term is not obj:
                        raise TypeError(
                            'Variant {variant} already has a contract term "{contract_term}" => {impl}. '
                            'It conflicts with {target}'.format(
                                variant=f'{self.variant_of.__sum_meta__.type.__module__}.{self.variant_of.__sum_meta__.type.__name__}::{self.name}',
                                contract_term=contract_term,
                                impl=f'{previous_term.__module__}:{previous_term.__name__}',
                                target=f'{obj.__module__}:{obj.__name__}'
                            )
                        )

                self.contract = self.contract._replace(**{contract_term: obj})

                # Re-calculate matches
                self.variant_of.__sum_meta__.matches = {
                    variant: {case: impl for case, impl in self.contract._asdict().items()}
                    for variant in self.variant_of.__sum_meta__.variants
                }
                scanner.configurator.sums.update_sum_type_registry(self.variant_of.__sum_meta__)

            info = self.venusian.attach(wrapped, callback, category='frameapp_namespace', depth=depth + 1)
            if info.scope == 'class':
                # if the decorator was attached to a method in a class, or
                # otherwise executed at class scope, we need to set an
                # 'attr' into the settings if one isn't already in there
                if settings.get('attr') is None:
                    settings['attr'] = wrapped.__name__
            return wrapped
        return wrapper

    def __repr__(self) -> str:
        return 'SumVariant(type={}.{}, name={}, value={})'.format(self.variant_of.__module__, self.variant_of.__name__, self.name, self.value)


class SumVariantInstance(SumVariant):
    """ A SumVariant that holds data associated with it
    """
    def __init__(self, data_args, data_kwargs, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.data = self.constructor(*data_args, **data_kwargs)

    def __eq__(self, other: SumVariant) -> bool:
        return super().__eq__(other) and self.data == getattr(other, 'data', None)

    __hash__ = SumVariant.__hash__

    def __repr__(self) -> str:
        return f'SumVariantInstance(type={self.variant_of.__module__}.{self.variant_of.__name__}, name={self.name}, value={self.value}, data={self.constructor})'


class SumTypeMetaData:
    __slots__ = ('type', 'variants', 'values', 'contract', 'matches')

    def __init__(self,
                 type,
                 variants: Dict[str, SumVariant],
                 values: Dict[str, str],
                 contract: NamedTuple,
                 matches: Dict[str, Dict[Any, Any]]) -> None:
        self.type = type
        self.variants = variants
        self.values = values
        self.contract = contract
        self.matches = matches


class SumTypeMetaclass(type):
    """ Metaclass object to be used with the actual SumType implementation.
    """
    def __new__(mcs, class_name: str, bases, attrs: Dict[str, Any]):
        """ This magic method is called when a new SumType class is being defined and parsed.
        """
        cls = type.__new__(mcs, class_name, bases, attrs)

        variants = {}
        variant_values = {}
        variant_constructors = get_type_hints(cls)

        # 0. Create a Contract, if specified
        # --------------------------------------------------
        try:
            contract_prototype = attrs['Contract']
        except KeyError:
            terms = {}
            contract_class: Type[NamedTuple] = NamedTuple(f'{class_name}Contract')
        else:
            terms = contract_prototype.__annotations__
            contract_class: Type[NamedTuple] = NamedTuple(f'{class_name}Contract', terms.items())

        contract = contract_class(**{k: None for k in terms.keys()})

        # 1. Populating variants from long-form definitions:
        #    class A(SumType):
        #        B[: type] = value
        # --------------------------------------------------
        for attr_name, value in attrs.items():
            # Populating variants from long-form definitions:
            # class A(SumType):
            #     B[: type] = value
            if not SUM_TYPE_VARIANT_NAME_RE.match(attr_name):
                continue

            if attr_name not in variant_constructors:
                raise ConfigurationError(f'SumType Variant "{cls.__module__}::{cls.__name__}::{attr_name}" '
                                         "must have a value constructor. "
                                         "You need to specify it as a type hint.")

            variant = SumVariant(variant_of=cls,
                                 name=attr_name,
                                 constructor=variant_constructors[attr_name],
                                 value=value,
                                 contract=contract)

            setattr(cls, attr_name, variant)
            variants[attr_name] = variant
            variant_values[value] = attr_name

        # 2. Populating variants from short-form definitions:
        #    class A(SumType):
        #        B: type
        # --------------------------------------------------
        # note that the value will be a lower-case version of the variant name
        for attr_name, constructor in variant_constructors.items():
            if not SUM_TYPE_VARIANT_NAME_RE.match(attr_name):
                continue
            if attr_name in variants:
                continue

            value = attr_name.lower()
            variant = SumVariant(variant_of=cls,
                                 name=attr_name,
                                 constructor=variant_constructors[attr_name],
                                 value=value,
                                 contract=contract)

            setattr(cls, attr_name, variant)
            variants[attr_name] = variant
            variant_values[value] = attr_name

        # 4. Let's create an instance of the contract and use it for replacement of the original
        # class definition. This will allow us to use SumType.Contract in bindings like:
        #
        # >>> @SomeSumType.VARIANT.bind(SomeSumType.Contract.contract_term)
        # >>> def some_term():
        # >>>     pass
        #
        # It is useful for "Find Usages" functionality of PyCharm
        # --------------------------------------------------------------------------------------
        c = contract_class(**{t: t for t in contract._fields})
        setattr(cls, 'Contract', c)

        # 5. Finalize
        # --------------------------------------------------
        cls.__sum_meta__ = SumTypeMetaData(
            type=cls,
            # set of SumType variants
            variants=variants,
            # dict of value => variant mappings
            values=variant_values,
            contract=contract,
            # dict of value => match instances.
            # Used by .match() for O(1) result retrieval
            matches={v: {} for v in variants}
        )
        return cls

    # Make the object iterable, similar to the standard enum.Enum
    def __iter__(cls) -> Iterator:
        return cls.__sum_meta__.variants.values().__iter__()


class SumType(metaclass=SumTypeMetaclass):
    __sum_meta__: SumTypeMetaData = None

    class Mismatch(Exception):
        pass

    class PatternError(Exception):
        pass

    @classmethod
    def values(cls) -> Set:
        return set(cls.__sum_meta__.values.keys())

    @classmethod
    def match(cls, value) -> SumVariant:
        """
        :rtype: dict or :class:`types.FunctionType`
        """
        variant = None
        for variant_name, variant_type in cls.__sum_meta__.variants.items():
            if variant_type.is_primitive_type():
                # We compare primitive types with equality matching
                if value == variant_type.value:
                    variant = variant_type
                    break
            else:
                # We compare non-primitive types with type checking
                if isinstance(value, variant_type.value):
                    variant = variant_type
                    break

        if variant is None:
            raise cls.Mismatch(
                'Variant value "{value}" is not a part of the type {type}: {values}'.format(
                    value=value,
                    type=cls.__sum_meta__.type,
                    values=u', '.join(['{val} => {var}'.format(val=val, var=var)
                                       for val, var in cls.__sum_meta__.values.items()])
                )
            )

        return variant

    @classmethod
    def inline_match(cls, **inline_cases):
        all_cases = set(cls.__sum_meta__.variants.keys())
        inline_cases = inline_cases.items()
        checked_cases = []
        for variant_name, fun in inline_cases:
            try:
                variant = cls.__sum_meta__.variants[variant_name]
            except KeyError:
                raise cls.PatternError(
                    'Variant {variant} does not belong to the type {type}'.format(
                        variant=str(variant_name),
                        type=cls.__sum_meta__.type,
                    )
                )
            all_cases.remove(variant.name)
            checked_cases.append((variant, fun))

        if all_cases:
            raise cls.PatternError(
                'Inline cases are not exhaustive.\n'
                'Here is the variant that is not matched: {variant} '.format(
                    variant=list(all_cases)[0]
                )
            )

        def matcher(value):
            for variant, fun in checked_cases:
                if variant.is_primitive_type():
                    if value == variant.value:
                        return fun
                else:
                    if isinstance(value, variant.value):
                        return fun

            raise cls.Mismatch(
                'Variant value "{value}" is not a part of the type {type}: {values}'.format(
                    value=value,
                    type=cls.__sum_meta__.type,
                    values=u', '.join(['{val} => {var}'.format(val=val, var=var)
                                       for val, var in cls.__sum_meta__.values.items()])
                )
            )
        return matcher

    def __init__(self) -> None:
        raise TypeError('SumType is a type, not an instance.')


class SumTypesConfigurator:

    def __init__(self) -> None:
        self.registry = OrderedDict()

    def update_sum_type_registry(self, sum_type_meta: SumTypeMetaData) -> None:
        self.registry[sum_type_meta.type] = sum_type_meta

    def check_sum_types_consistency(self, namespace) -> None:
        for obj_id, sum_type_meta in self.registry.items():
            for variant_name, variant in sum_type_meta.variants.items():
                for contract_term, implementation in variant.contract._asdict().items():
                    if implementation is None:
                        raise ConfigurationError(
                            'Contract term "{contract_term}" of the sum type {type} is not complete. '
                            'Here is the missing variant: {variant} '
                            .format(
                                contract_term=contract_term,
                                type=f'{sum_type_meta.type.__module__}.{sum_type_meta.type.__name__}',
                                variant=f'{sum_type_meta.type.__module__}.{sum_type_meta.type.__name__}::{variant_name}'
                            )
                        )
            log.debug(f'Checked for {sum_type_meta.type.__module__}.{sum_type_meta.type.__name__}. '
                      f'Registered contract: {sum_type_meta.contract}')



""" This module is initially taken from Solo

https://github.com/avanov/solo/blob/bf44c527dbe48256d2bd3da463eceeb78d05a38d/solo/configurator/config/predicates.py
"""
import operator
from typing import Optional

import jsonschema
from django.http import HttpRequest
from pkg_resources.extern.packaging.version import Version
from pkg_resources import parse_version
from rest_framework.request import Request

from ..schemas import json_schema
from .util import as_sorted_tuple


class RequestMethodPredicate:
    def __init__(self, val, config, raises: Optional[Exception] = None) -> None:
        """ Predicates are constructed at ``solo.configurator.config.util.PredicateList.make()``

        :param val: value passed to http_endpoint/http_defaults
        :param config:
        """
        request_method = as_sorted_tuple(val)
        if 'GET' in request_method and 'HEAD' not in request_method:
            # GET implies HEAD too
            request_method = as_sorted_tuple(request_method + ('HEAD',))
        self.val = request_method
        self.raises = raises

    def text(self) -> str:
        return 'request_method<%s>' % (','.join(self.val))

    phash = text
    __repr__ = text

    def __call__(self, context: Optional, request: HttpRequest) -> bool:
        """
        :param context: at the moment context may be only None
        :type context: None
        :param: request: Django request object
        :type request: :class:`django.http.HttpRequest`
        """
        return request.method in self.val


class ApiVersionPredicate:
    OPERATORS = {
        '>': operator.gt,
        '<': operator.lt,
        '==': operator.eq,
        '>=': operator.ge,
        '<=': operator.le
    }

    def __init__(self, val, config, raises: Optional[Exception] = None) -> None:
        """
        :param val: value passed to view_config/view_defaults
        :param config:
        """
        api_version = as_sorted_tuple(val)
        self.val = api_version
        self.raises = None

    def text(self) -> str:
        return f'api_version<{self.val}>'

    phash = text
    __repr__ = text

    def __call__(self, context: Optional, request: HttpRequest) -> bool:
        """
        :param context: at the moment context may be only None
        :type context: None
        :param: request: Django request object
        :type request: :class:`django.http.HttpRequest`
        """
        for allowed_api_pattern in self.val:
            if self.match_api_version(request.API_VERSION, allowed_api_pattern):
                return True
        return False

    def match_api_version(self, request_version: Version, allowed_version) -> bool:
        """
        :param allowed_version: may be represented in following forms:
            1. ``VERSION``
            2. ``==VERSION`` (the same as above)
            3. ``>VERSION``
            4. ``<VERSION``
            5. ``>=VERSION``
            6. ``<=Version``
            7. Comma-separated list of 1-7 evaluated as AND
        :return: :raise ValueError:
        """
        distinct_versions = {version.strip() for version in allowed_version.split(',')}
        for distinct_version in distinct_versions:
            operation = self.OPERATORS.get(distinct_version[:2])
            if operation:
                # prepare cases #2, #5, #6
                compare_with = distinct_version[2:]
            else:
                operation = self.OPERATORS.get(distinct_version[0])
                if operation:
                    # prepare cases #3, #4
                    compare_with = distinct_version[1:]
                else:
                    # prepare case #1
                    compare_with = distinct_version
                    operation = self.OPERATORS['==']

            # evaluate the case
            matched = operation(request_version, parse_version(compare_with))
            return matched
        return True


class OutputSerializerPredicate:
    def __init__(self, val, config, raises: Optional[Exception] = None) -> None:
        """ Predicates are constructed at ``solo.configurator.config.util.PredicateList.make()``

        :param val: value passed to http_endpoint/http_defaults
        :param config:
        """

        self.val = val
        self.raises = raises

    def text(self) -> str:
        return f'output_serializer<{self.val.__name__}>'

    phash = text
    __repr__ = text

    def __call__(self, context: Optional, request: HttpRequest) -> bool:
        """
        :param context: at the moment context may be only None
        :type context: None
        :param: request: Django request object
        :type request: :class:`django.http.HttpRequest`
        """
        return True


class InputSchemaPredicate:
    def __init__(self, val, config, raises: Optional[Exception] = None) -> None:
        """ Predicates are constructed at ``solo.configurator.config.util.PredicateList.make()``

        :param val: value passed to http_endpoint/http_defaults
        :param config:
        """
        if val.endswith('.json'):
            self.val_file = val
            self.val = json_schema(val)
        else:
            self.val_file = None
            self.val = val

        self.raises = raises

    def text(self) -> str:
        if self.val_file:
            rv = f'input_schema<{self.val}>'
        else:
            rv = f'input_schema<...>'
        return rv

    phash = text
    __repr__ = text

    def __call__(self, context: Optional, request: Request) -> bool:
        jsonschema.validate(request.data, self.val)
        return True


class OutputSchemaPredicate:
    def __init__(self, val, config, raises: Optional[Exception] = None) -> None:
        """ Predicates are constructed at ``solo.configurator.config.util.PredicateList.make()``

        :param val: value passed to http_endpoint/http_defaults
        :param config:
        """

        self.val = val
        self.raises = raises

    def text(self) -> str:
        if self.val.endswith('.json'):
            rv = f'output_schema<{self.val}>'
        else:
            rv = f'output_schema<...>'
        return rv

    phash = text
    __repr__ = text

    def __call__(self, context: Optional, request: HttpRequest) -> bool:
        """
        :param context: at the moment context may be only None
        :type context: None
        :param: request: Django request object
        :type request: :class:`django.http.HttpRequest`
        """
        return True



""" This module is initially taken from Solo

https://github.com/avanov/solo/blob/bf44c527dbe48256d2bd3da463eceeb78d05a38d/solo/configurator/config/rendering.py
"""
from typing import Dict, Any, TypeVar, List, Generator
from enum import Enum
from decimal import Decimal
from collections import KeysView, ValuesView, ItemsView
from datetime import datetime, tzinfo

from pytz.tzinfo import BaseTzInfo
from wrapt import ObjectProxy
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpRequest, HttpResponse
from rest_framework.utils import encoders


JsonPayload = TypeVar('JsonApiPayload', Dict[str, Any], List[Dict[str, Any]])


class JsonRendererFactory:
    def __init__(self, name: str) -> None:
        self.name = name

    def __call__(self, request: HttpRequest, view_response: JsonPayload) -> HttpResponse:
        return HttpResponse(status=200,
                            content=encode_json(view_response).encode('utf-8'),
                            content_type='application/json',
                            charset='utf-8')


class StringRendererFactory:
    def __init__(self, name: str) -> None:
        self.name = name

    def __call__(self, request: HttpRequest, view_response: Any):
        HttpResponse(status=200,
                     content=str(view_response).encode('utf-8'),
                     content_type='text/plain',
                     charset='utf-8')


BUILTIN_RENDERERS = {
    'json': JsonRendererFactory,
    'string': StringRendererFactory,
}


class RenderersConfigurator:

    def __init__(self) -> None:
        self.registry = {}
        for name, renderer in BUILTIN_RENDERERS.items():
            self.add_renderer(name, renderer)

    def add_renderer(self, name: str, factory):
        self.registry[name] = factory

    def get_renderer(self, name: str):
        try:
            template_suffix = name.rindex(".")
        except ValueError:
            # period is not found
            renderer_name = name
        else:
            renderer_name = name[template_suffix:]

        try:
            rv = self.registry[renderer_name](name)
        except KeyError:
            raise ValueError(f'No such renderer factory "{renderer_name}"')
        return rv


class ExtendedJSONEncoder(encoders.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if hasattr(obj, '__json__'):
            return obj.__json__()
        elif isinstance(obj, (BaseTzInfo, tzinfo)):
            return str(obj)
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, ObjectProxy):
            return obj.__wrapped__

        return super().default(obj)


class ExtendedDjangoJSONEncoder(DjangoJSONEncoder):
    def default(self, o):
        """
        Check that milliseconds/microseconds are ignored when encoding datetime
        objects (we don't save them in the database, but some endpoints end up
        making them - and milliseconds break Jon's iOS JSON Decoder).
        """
        if hasattr(o, '__json__'):
            return o.__json__()
        elif isinstance(o, datetime) and o.microsecond:
            return o.replace(microsecond=0)
        # http://stackoverflow.com/questions/1960516/python-json-serialize-a-decimal-object
        elif isinstance(o, Decimal):
            return float(o)
        elif o.__class__.__name__ is '__proxy__':
            return str(o)
        elif isinstance(o, (KeysView, ValuesView, ItemsView, Generator)):
            return list(o)
        elif isinstance(o, Enum):
            return o.value

        return super().default(o)


encode_json = ExtendedDjangoJSONEncoder().encode



import logging
from typing import List, Dict, Union, Optional

from django.http.request import HttpRequest as DjangoRequest
from django.http.response import HttpResponse as DjangoResponse
from django.http import Http404 as HTTPNotFound
from rest_framework.request import Request as DRFRequest
from rest_framework.response import Response as DRFResponse

from frameapp.configurator.routes import ViewVariant
from frameapp.configurator.sums import SumType


log = logging.getLogger(__name__)


HttpRequest = Union[DjangoRequest, DRFRequest]
HttpResponse = Union[DjangoResponse, DRFResponse]


class PredicatedHandler:
    """ Wrapper object around actual view handlers that checks predicates during the request
    and processes results returned from view handlers during the response.
    """
    __slots__ = ['rules', 'view_variants', 'csrf_exempt']

    def __init__(self, rules: Dict[str, SumType], view_variants: List[ViewVariant]) -> None:
        self.view_variants = view_variants
        # Note that the entire PredicatedHandler will be CSRF-exempt if at least one handler is exempt.
        # This is not the ideal solution, yet it allows us to encapsulate the logic inside this object
        # instead of implementing csrf check on individual handlers in the middleware, when predicates are matched
        # and the final handler is known.
        # Most of the time it's what we need anyway, because we don't put non-api views inside DRF APIView subclasses.
        self.csrf_exempt = any(getattr(v.handler, 'csrf_exempt', False) for v in view_variants)
        self.rules = rules

    def match_predicates(self, request: HttpRequest) -> Optional[ViewVariant]:
        for view_variant in self.view_variants:
            for predicate in view_variant.predicates:
                if not predicate(None, request):
                    log.debug(f'Predicate {predicate} failed for {request.method} {request.path}')
                    break
            else:
                return view_variant
        return None

    def __call__(self, request: HttpRequest, *route_args, **route_kwargs) -> HttpResponse:
        """ Try to resolve predicates and call a view handler on success.
        """
        # here predicate is an instance object
        matched_view_variant = self.match_predicates(request)
        if matched_view_variant:
            # All predicates match, proceed to view handler invocation
            log.debug(f'{request.method} {request.path} will be handled by {matched_view_variant.handler}')
            handler = matched_view_variant.handler
            context = {}
            rules = self.rules
            for k, v in route_kwargs.items():
                if k in rules:  # match SumType's case
                    context[k] = self.rules[k].match(v)
                else:  # regular value assignment
                    context[k] = v

            response = handler(request, *route_args, **route_kwargs)
            #     # else:
            #     #     # Handler is a Pyramid-like class view.
            #     #     handler = getattr(handler(request, context), matched_view_variant.attr)
            #     #     response = handler()
            # else:
            #     # handler is a simple callable
            #     response = handler(request, context, *route_args, **route_kwargs)

            if isinstance(response, DjangoResponse):
                # Do not process standard responses
                final_response = response
            else:
                renderer = matched_view_variant.renderer
                final_response = renderer(request, response)

            return final_response

        log.debug(f'All predicates have failed for {request.method} {request.path}')
        raise HTTPNotFound()



import json
from typing import Dict, Callable, List, Optional, Iterable, Set, Any

from .configurator.renderers import ExtendedJSONEncoder


encode_json = ExtendedJSONEncoder().encode


class CompactSerializable:
    """ Whenever you inherit from this object, make sure that you define `__slots__` in your child classes.
    """
    __slots__ = tuple()
    IN_MODIFIERS: Dict[str, Callable] = {}
    OUT_MODIFIERS: Dict[str, Callable] = {
        'str': str,
        'json': json.dumps,
    }
    DEFAULT_IN_MODIFIERS: Dict[str, List[str]] = {}
    DEFAULT_OUT_MODIFIERS: Dict[str, List[str]] = {}

    def as_dict(self,
                fields: Optional[Iterable[str]] = None,
                exclude: Optional[Set[str]] = None) -> Dict[str, Any]:
        """ Returns a dictionary representation of itself, adhering provided fields and modifiers.
        """
        if exclude is None:
            exclude = set()
        if not fields:
            fields = (k for k in self.__slots__ if k not in exclude)

        rv = {}
        for f in fields:
            split = f.split('|')
            field, modifiers = split[0], split[1:]
            field = field.strip()
            v = getattr(self, field)
            for m in modifiers:
                v = self.apply_modifier(m.strip(), v)
            else:
                default_modifiers = self.DEFAULT_OUT_MODIFIERS.get(field, [])
                for m in default_modifiers:
                    v = self.apply_modifier(m, v)
            rv[field] = v
        return rv

    def apply_modifier(self, modifier: str, value: Any) -> Any:
        if modifier.startswith('.'):
            # applying a method of the value
            value = getattr(value, modifier[1:])()
        else:
            value = self.OUT_MODIFIERS[modifier](value)
        return value

    def __str__(self) -> str:
        """ Serialize the object by passing it to ``str()``.
        """
        return encode_json(self.as_dict())

    def __bytes__(self) -> bytes:
        return bytes(self.__str__(), 'utf-8')



""" This module is initially taken from Solo

https://github.com/avanov/solo/blob/bf44c527dbe48256d2bd3da463eceeb78d05a38d/solo/configurator/config/views.py
"""
import inspect
import logging
from collections import OrderedDict
from typing import Optional

from django.views.generic.base import View as DjangoGenericView

from . import predicates as default_predicates
from ..util import viewdefaults
from .routes import ViewMeta
from .util import PredicateList
from ..exceptions import ConfigurationError


log = logging.getLogger(__name__)


class ViewsConfigurator:

    def __init__(self) -> None:
        self.registry = OrderedDict()
        self.predicates = PredicateList()
        self.add_default_view_predicates()

    @viewdefaults
    def add_view(self,
                 view=None,
                 route_name: Optional[str] = None,
                 request_method=None,
                 attr=None,
                 decorator=None,
                 renderer=None,
                 **predicates) -> ViewMeta:
        """

        :param view: callable
        :param route_name:
        :param request_method:
        :type request_method: str or tuple
        :param attr:
          This knob is most useful when the view definition is a class.

          The view machinery defaults to using the ``__call__`` method
          of the :term:`view callable` (or the function itself, if the
          view callable is a function) to obtain a response.  The
          ``attr`` value allows you to vary the method attribute used
          to obtain the response.  For example, if your view was a
          class, and the class has a method named ``index`` and you
          wanted to use this method instead of the class' ``__call__``
          method to return the response, you'd say ``attr="index"`` in the
          view configuration for the view.
        :type attr: str
        :param decorator:
        :param renderer:
        :param predicates: Pass a key/value pair here to use a third-party predicate
                           registered via
                           :meth:`solo.configurator.config.Configurator.views.add_view_predicate`.
                           More than one key/value pair can be used at the same time. See
                           :ref:`view_and_route_predicates` for more information about
                           third-party predicates.
        :return: :raise ConfigurationError:
        """
        from rest_framework.viewsets import ViewSetMixin

        # Prepare view object
        # -------------------------------------
        is_django_generic_view = issubclass(view, DjangoGenericView)
        is_drf_model_viewset = issubclass(view, ViewSetMixin)

        if inspect.isclass(view):
            if attr is None:
                attr = '__call__'
            if not issubclass(view, DjangoGenericView) and not hasattr(view, attr):
                raise ConfigurationError(f"View {view} is registered as a callable, but didn't define its __call__ method.")

        # Add decorators
        # -----------------------------------------------
        def combine(*decorators):
            def decorated(view_callable):
                # reversed() is allows a more natural ordering in the api
                for decorator in reversed(decorators):
                    view_callable = decorator(view_callable)
                return view_callable
            return decorated

        if isinstance(decorator, (tuple, list)):
            decorator = combine(*decorator)

        # Register predicates
        # -------------------------------------
        if request_method is None:
            request_method = ('GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'PATCH', 'OPTIONS')
        pvals = predicates.copy()
        pvals.update(dict(request_method=request_method,))
        predlist = self.get_predlist('view')
        _weight_, preds, _phash_ = predlist.make(self, **pvals)

        # Renderers
        # -------------------------------------
        if renderer is None:
            renderer = 'string'

        # Done
        # -------------------------------------
        view_item = ViewMeta(route_name=route_name,
                             registered_view=view,
                             actual_view_handler=None,
                             decorator=decorator,
                             attr=attr,
                             renderer=renderer,
                             predicates=preds,
                             is_django_generic_view=is_django_generic_view,
                             is_drf_model_viewset=is_drf_model_viewset)

        log.debug(f'View added: {view_item}')
        return view_item

    def get_predlist(self, name: str):
        """ This is a stub method that simply has the same signature as pyramid's version,
        but does nothing but returning ``self.predicates``
        """
        return self.predicates

    def add_view_predicate(self, name, factory, weighs_more_than=None,
                           weighs_less_than=None):
        """
        Adds a view predicate factory.  The associated view predicate can
        later be named as a keyword argument to
        :meth:`solo.configurator.config.Configurator.views.add_view` in the
        ``predicates`` anonymous keyword argument dictionary.

        ``name`` should be the name of the predicate.  It must be a valid
        Python identifier (it will be used as a keyword argument to
        ``add_view`` by others).

        ``factory`` should be a :term:`predicate factory` or :term:`dotted
        Python name` which refers to a predicate factory.

        See :ref:`view_and_route_predicates` for more information.
        """
        self._add_predicate(
            'view',
            name,
            factory,
            weighs_more_than=weighs_more_than,
            weighs_less_than=weighs_less_than
        )

    def add_default_view_predicates(self):
        p = default_predicates
        for name, factory in (('request_method', p.RequestMethodPredicate),
                              ('api_version', p.ApiVersionPredicate),
                              ('output_serializer', p.OutputSerializerPredicate),
                              ('input_schema', p.InputSchemaPredicate),
                              ('output_schema', p.OutputSchemaPredicate),):
            self.add_view_predicate(name, factory)

    def _add_predicate(self, type: str, name: str, factory, weighs_more_than=None, weighs_less_than=None):
        """ This method is a highly simplified equivalent to what you can find in Pyramid.

        :param type: may be only 'view' at the moment
        :param name: valid python identifier string.
        :param weighs_more_than: not used at the moment
        :param weighs_less_than: not used at the moment
        """
        predlist = self.get_predlist(type)
        predlist.add(name, factory, weighs_more_than=weighs_more_than,
                     weighs_less_than=weighs_less_than)



import os
import logging
from collections import OrderedDict
from typing import Optional, Dict, List, NamedTuple, Any, Callable, Union, Sequence, Set, Tuple

from .sums import SumType
from ..util import viewdefaults
from ..exceptions import ConfigurationError


log = logging.getLogger(__name__)


class RoutesConfigurator:
    def __init__(self, route_prefix: str = '/', namespace: str = 'frameapp_namespace') -> None:
        self.registry: Dict[str, Dict[str, Route]] = OrderedDict()
        self.route_prefix = route_prefix
        self.namespace = namespace
        self.change_namespace(namespace)

    def change_route_prefix(self, prefix: str) -> str:
        old_prefix = self.route_prefix
        self.route_prefix = prefix
        return old_prefix

    def change_namespace(self, new: str) -> str:
        old = self.namespace
        self.namespace = new
        self.registry.setdefault(new, OrderedDict())
        return old

    def add_route(self, name: str, pattern: str, rules: Optional[Dict[str, SumType]] = None, extra_kwargs=None) -> None:
        pattern = os.path.join(self.route_prefix.rstrip('/'), pattern.lstrip('/'))
        if not pattern:
            pattern = '/'

        if rules is None:
            rules = {}

        if name in self.registry[self.namespace]:
            raise ConfigurationError(f'Route named "{name}" is already registered in the namespace "{self.namespace}"')

        log.debug(f'Registering global route "{pattern}" with the local name "{name}" in the "{self.namespace}" namespace')
        self.registry[self.namespace][name] = Route(name=name,
                                                    pattern=pattern,
                                                    rules=rules,
                                                    extra_kwargs=extra_kwargs,
                                                    view_metas=[],
                                                    # This field might be updated later
                                                    schemas=OrderedDict())

    @viewdefaults
    def add_route_schema(self,
                         view,
                         route_name: str,
                         request_method: Union[None, str, Sequence] = None,
                         attr: Optional[str] = None,
                         input_serializer: Any = None,
                         output_serializer: Any = None,
                         input_schema: Any = None,
                         output_schema: Any = None,
                         api_version: Any = None,
                         **extra_kwargs) -> None:

        route = self.registry[self.namespace][route_name]

        if any((input_serializer, output_serializer, input_schema, output_schema, api_version)):
            if request_method is None:
                request_method = ('GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'PATCH', 'OPTIONS')
            else:
                if isinstance(request_method, list):
                    request_method = tuple(request_method)
                else:
                    request_method = (request_method, )

            sig = SchemaIdentifier(request_method=request_method,
                                   route_name=route_name,
                                   api_version=api_version)
            if sig in route.schemas:
                raise ConfigurationError(f'Attempt to rewrite an existing schema for {sig}')

            if attr:
                doc = getattr(view, attr).__doc__
            else:
                doc = view.__doc__

            doc = doc or 'DOCSTRING NOT SPECIFIED'

            log.debug(f'Adding schema definition for {sig}')
            route.schemas[sig] = SchemaDefinition(
                input_serializer=input_serializer,
                output_serializer=output_serializer,
                input_schema=input_schema,
                output_schema=output_schema,
                doc=doc
            )

    def check_routes_consistency(self, namespace):
        log.debug(f'Checking routes consistency for namespace "{namespace}"...')
        for route_name, route in self.registry[namespace].items():
            view_metas = route.view_metas
            if not view_metas:
                raise ConfigurationError(
                    'Route name "{name}" is not associated with a view handler in the "{namespace}" namespace.'.format(
                        name=route_name,
                        namespace=namespace
                    )
                )
            for view_item in view_metas:
                if view_item.registered_view is None:
                    raise ConfigurationError(
                        'Route name "{name}" is not associated with a view handler in the "{namespace}" namespace.'.format(
                            name=route_name,
                            namespace=namespace
                        )
                    )


class ViewMeta(NamedTuple):
    route_name: str
    registered_view: Any
    """ View object that was registered by a marker
    """
    actual_view_handler: Optional[Callable]
    """ View object that is inferred to handle requests; inference is based on the kind of `registered_view`
    """
    decorator: Optional[Callable]
    attr: Optional[str]
    renderer: Any
    predicates: Any
    is_django_generic_view: bool
    is_drf_model_viewset: bool


class SchemaIdentifier(NamedTuple):
    request_method: Tuple[str]
    route_name: str
    api_version: str


class SchemaDefinition(NamedTuple):
    input_serializer: Optional[Any]
    output_serializer: Optional[Any]
    input_schema: Optional[Any]
    output_schema: Optional[Any]
    doc: Optional[str]


class Route(NamedTuple):
    name: str
    pattern: str
    rules: Any
    extra_kwargs: Optional[Any]
    view_metas: List[ViewMeta]
    schemas: Dict[SchemaIdentifier, SchemaDefinition]


class ViewVariant(NamedTuple):
    route_name: str
    registered_view: Any
    """ View object that was registered by a marker
    """
    handler: Callable
    """ View object that is inferred to handle requests; inference is based on the kind of `registered_view`
    """
    attr: Optional[str]
    renderer: Any
    predicates: Any



""" Initially taken from Solo

https://github.com/avanov/solo/blob/86695ede6f69a9a162943a4db03dd412ee3419c6/solo/configurator/url.py
"""
import re
import logging
from typing import Dict, List, Tuple, NamedTuple, Any, Optional, Type

from django.http import HttpRequest
from django.urls import URLPattern
from django.urls.resolvers import RegexPattern
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet
from rest_framework import routers

from frameapp.exceptions import ConfigurationError
from frameapp.configurator import Configurator
from frameapp.configurator.sums import SumType
from frameapp.configurator.routes import ViewVariant
from frameapp.util import maybe_dotted
from frameapp.ext.django_integration.view import PredicatedHandler


log = logging.getLogger(__name__)


# A replacement marker in a pattern must begin with an uppercase or
# lowercase ASCII letter or an underscore, and can be composed only
# of uppercase or lowercase ASCII letters, underscores, and numbers.
# For example: a, a_b, and b9 are all valid replacement marker names, but 0a is not.
ROUTE_PATTERN_OPEN_BRACES_RE = re.compile('(?P<start_brace>\{).*')
ROUTE_PATTERN_CLOSING_BRACES_RE = re.compile('\}.*')
EMBEDDED_SUM_TYPE_RE = re.compile('\<(?P<sum_type>[a-zA-Z0-9_.:]+)\>')


class Dispatcher(NamedTuple):
    route_namespace: str
    route_name: str
    route_pattern: str
    route_rules: Any
    route_schemas: Any
    route_extra_kwargs: Optional[Any]
    view_variants: List[ViewVariant]


def _extract_braces_expression(line: str, starting_braces_re, open_braces_re, closing_braces_re):
    """ This function is taken from Plim package: https://pypi.python.org/pypi/Plim/

    :param line: may be empty
    :param starting_braces_re:
    :param open_braces_re:
    :param closing_braces_re:
    """
    match = starting_braces_re.match(line)
    if not match:
        return None

    open_brace = match.group('start_brace')
    buf = [open_brace]
    tail = line[len(open_brace):]
    braces_counter = 1

    while tail:
        current_char = tail[0]
        if closing_braces_re.match(current_char):
            braces_counter -= 1
            buf.append(current_char)
            if braces_counter:
                tail = tail[1:]
                continue
            return ''.join(buf), tail[1:]

        if open_braces_re.match(current_char):
            braces_counter += 1
            buf.append(current_char)
            tail = tail[1:]
            continue

        buf.append(current_char)
        tail = tail[1:]
    raise Exception(f"Unexpected end of a route pattern: {line}")


extract_pattern = (lambda line: _extract_braces_expression(
    line,
    ROUTE_PATTERN_OPEN_BRACES_RE,
    ROUTE_PATTERN_OPEN_BRACES_RE,
    ROUTE_PATTERN_CLOSING_BRACES_RE
))


def normalize_route_pattern(pattern: str) -> Tuple[str, Dict[str, SumType]]:
    buf = []
    rules = {}
    while pattern:
        result = extract_pattern(pattern)
        if result:
            extracted, pattern = result
            if ':' in extracted:
                # Remove braces from the extracted result "{pattern[:rule]}"
                pattern_name, rule = extracted[1:-1].split(':', 1)
                embedded_sum_type = EMBEDDED_SUM_TYPE_RE.match(rule)
                if embedded_sum_type:
                    sum_type = maybe_dotted(embedded_sum_type.group('sum_type'))
                    rules[pattern_name] = sum_type
                    url_part = '{{{}}}'.format(pattern_name)
                else:
                    url_part = extracted
            else:
                url_part = extracted

            buf.append(url_part)
            continue

        buf.append(pattern[0])
        pattern = pattern[1:]

    # Parsing is done. Now join everything together
    buf = ''.join(buf)
    return buf, rules


_django_rule_format = (lambda match_group_name, rule: f"(?P<{match_group_name}>{rule})")
_solo_rule_format = (lambda match_group_name, rule: f"{{{match_group_name}:{rule}}}")


def complete_route_pattern(pattern: str, rules: Dict[str, Type[SumType]], rule_format=_solo_rule_format) -> str:
    """
    :param pattern: URL pattern
    """
    buf = []
    while pattern:
        result = extract_pattern(pattern)
        if result:
            result, pattern = result
            # Remove braces from the result "{pattern[:rule]}"
            result = result[1:-1]
            if ':' in result:
                # pattern in a "pattern_name:rule" form
                match_group_name, rule = result.split(':', 1)
            else:
                # pattern in a "pattern_name" form
                match_group_name = result
                rule = rules.get(match_group_name)
                if not rule:
                    # Use default pattern
                    rule = '[^/]+'
                elif issubclass(rule, SumType):
                    # Compose SumType's regex
                    rule = '(?:{})'.format('|'.join([str(v) for v in rule.values()]))

            result = rule_format(match_group_name, rule)
            buf.append(result)
            continue

        buf.append(pattern[0])
        pattern = pattern[1:]

    # Parsing is done. Now join everything together
    buf = ''.join(buf)
    return buf


def create_django_route(dispatcher: Dispatcher) -> List[URLPattern]:
    view_variants = []
    viewset_variants = []
    for view_variant in dispatcher.view_variants:
        if isinstance(view_variant.handler, DRFViewMixinWrapper):
            vv = viewset_variants
        else:
            vv = view_variants
        vv.append(view_variant)

    rv = []
    # Prepare URLPatterns similar to http://www.django-rest-framework.org/api-guide/routers/#usage
    if view_variants:
        django_route_name = f'{dispatcher.route_namespace}.{dispatcher.route_name}'
        pattern = dispatcher.route_pattern.lstrip('/')  # removes django warnings
        pattern = complete_route_pattern(pattern, dispatcher.route_rules, _django_rule_format)
        regex_pattern = f'^{pattern}$'
        callback = PredicatedHandler(dispatcher.route_rules, view_variants)
        log.debug(f'Creating Django URL "{regex_pattern}" as the handler named "{dispatcher.route_name}" in the namespace "{dispatcher.route_namespace}".')
        rv.append(URLPattern(RegexPattern(regex_pattern, is_endpoint=True), callback, dispatcher.route_extra_kwargs, django_route_name))

    if viewset_variants:
        django_route_name = f'{dispatcher.route_namespace}.{dispatcher.route_name}-drf_viewset'
        pattern = dispatcher.route_pattern.lstrip('/')  # removes django warnings
        pattern = complete_route_pattern(pattern, dispatcher.route_rules, _django_rule_format)
        drf_pattern = '(?P<__frameapp_dynamic_viewset__>/(?P<pk>[^/.]+))?'
        regex_pattern = f'^{pattern}{drf_pattern}$'
        callback = PredicatedHandler(dispatcher.route_rules, viewset_variants)
        log.debug(f'Creating DRF URL "{regex_pattern}" as the handler named "{dispatcher.route_name}" in the namespace "{dispatcher.route_namespace}".')
        rv.append(URLPattern(RegexPattern(regex_pattern, is_endpoint=True), callback, dispatcher.route_extra_kwargs, django_route_name))
    return rv


def generate_api_docs(configurator: Configurator):
    return ''


def django_url_patterns(namespace: str, configurator: Configurator) -> List[URLPattern]:
    """ Generates Django URLs from registered routes
    """
    application_routes = configurator.routes.registry[namespace]
    rv = []
    dispatchers = []
    for route in application_routes.values():
        dispatcher = Dispatcher(
            route_namespace=namespace,
            route_name=route.name,
            route_pattern=route.pattern,
            route_rules=route.rules,
            route_schemas=route.schemas,
            route_extra_kwargs=route.extra_kwargs,
            view_variants=[]
        )
        for view_meta in route.view_metas:
            if issubclass(view_meta.registered_view, GenericViewSet):
                handler = DRFViewMixinWrapper(view_meta.registered_view)
            elif issubclass(view_meta.registered_view, APIView):
                handler = DRFAPIViewWrapper(view_meta.registered_view, view_meta.attr)
            else:
                raise ConfigurationError(f'Unknown type of view: {view_meta.registered_view}')

            if view_meta.decorator:
                # apply decorators
                handler = view_meta.decorator(handler)

            view_variant = ViewVariant(
                route_name=view_meta.route_name,
                registered_view=view_meta.registered_view,
                handler=handler,
                attr=view_meta.attr,
                renderer=view_meta.renderer,
                predicates=view_meta.predicates
            )
            dispatcher.view_variants.append(view_variant)

        dispatchers.append(dispatcher)

    for dispatcher in dispatchers:
        rv += create_django_route(dispatcher=dispatcher)

    return rv


class DRFViewMixinWrapper:
    """ Wrap DRF ViewMixin into a callable object that dispatches requests either to a collection or a item view.
    """
    def __init__(self, view_cls: Type[GenericViewSet]) -> None:
        router = routers.SimpleRouter(trailing_slash=False)
        router.register('', view_cls)
        urls: Tuple[URLPattern, URLPattern] = router.urls
        collection, item = urls
        self.collection_handler = collection.callback
        self.item_handler = item.callback
        self.csrf_exempt = any([
            getattr(self.collection_handler, 'csrf_exempt', False),
            getattr(self.item_handler, 'csrf_exempt', False),
        ])

    def __call__(self, request: HttpRequest, *args, **kwargs):
        item_suffix = kwargs.pop('__frameapp_dynamic_viewset__', '')
        if item_suffix:
            handler = self.item_handler
        else:
            handler = self.collection_handler
        return handler(request, *args, **kwargs)

    def __repr__(self) -> str:
        return f'DRFViewMixinWrapper(handler=[{self.collection_handler}, {self.item_handler}])'

    @staticmethod
    def drf_frameapp_dispatch_patched_version(self: APIView, __frameapp_view_attr__: str, request, *args, **kwargs):
        """ This portion of code is taken from DRF
        https://github.com/encode/django-rest-framework/blob/79be20a7c68e7c90dd4d5d23a9e6ee08b5f586ae/rest_framework/views.py#L465

        and then modified in order to support multiple handlers for the same request.
        __frameapp_view_attr__ will always contain a name to a method whose predicates matched the request.

        Original docstring:

        `.dispatch()` is pretty much the same as Django's regular dispatch,
        but with extra hooks for startup, finalize, and exception handling.
        """
        handler = getattr(self, __frameapp_view_attr__)
        log.debug(f'Entering patched DRF resolver with target method "{handler}"')

        self.args = args
        self.kwargs = kwargs
        request = self.initialize_request(request, *args, **kwargs)
        self.request = request
        self.headers = self.default_response_headers  # deprecate?

        try:
            self.initial(request, *args, **kwargs)
            response = handler(request, *args, **kwargs)
        except Exception as exc:
            response = self.handle_exception(exc)

        self.response = self.finalize_response(request, response, *args, **kwargs)
        return self.response


class DRFAPIViewWrapper:
    """ Wrap DRF APIView into a callable object that dispatches requests based on view attribute name.
    """
    def __init__(self, view_cls: Type[APIView], view_attr: str) -> None:
        self.attr = view_attr
        view_cls.dispatch = _drf_frameapp_dispatch_patched_version  # this has a global side-effect, not good enough

        self.handler = view_cls.as_view()
        self.csrf_exempt = getattr(self.handler, 'csrf_exempt', False)
        # Let's monkey-patch this view class with a b/w compatible .dispatch() method that understands
        # how to dispatch to predicated methods, and then let's generate an actual view handler.
        self.handler.dispatch = _drf_frameapp_dispatch_patched_version

    def __call__(self, request, *args, **kwargs):
        # DRF GenericViewSet is similar to DRF ModelViewSet, yet we have an extra step here.
        # The handler will call a patched version of DRFs .dispatch() method that will pass data to the correct method.

        # Note however, that by the time DRF is in action, we have matched all the required predicates.
        # Best of the two worlds!
        return self.handler(self.attr, request, *args, **kwargs)


# Note: DRF API Views are made CSRF exempt from within `as_view` as to prevent
# accidental removal of this exemption in cases where `dispatch` needs to
# be overridden.
def _drf_frameapp_dispatch_patched_version(self: APIView, __frameapp_view_attr__: str, request, *args, **kwargs):
    """ This portion of code is taken from DRF
    https://github.com/encode/django-rest-framework/blob/79be20a7c68e7c90dd4d5d23a9e6ee08b5f586ae/rest_framework/views.py#L465

    and then modified in order to support multiple handlers for the same request.
    __frameapp_view_attr__ will always contain a name to a method whose predicates matched the request.

    Original docstring:

    `.dispatch()` is pretty much the same as Django's regular dispatch,
    but with extra hooks for startup, finalize, and exception handling.
    """
    handler = getattr(self, __frameapp_view_attr__)
    log.debug(f'Entering patched DRF resolver with target method "{handler}"')

    self.args = args
    self.kwargs = kwargs
    request = self.initialize_request(request, *args, **kwargs)
    self.request = request
    self.headers = self.default_response_headers  # deprecate?

    try:
        self.initial(request, *args, **kwargs)
        response = handler(request, *args, **kwargs)
    except Exception as exc:
        response = self.handle_exception(exc)

    self.response = self.finalize_response(request, response, *args, **kwargs)
    return self.response



import inspect
import functools

import pkg_resources

from .exceptions import ConfigurationError, CyclicDependencyError


def is_nonstr_iter(v):
    if isinstance(v, str):
        return False
    return hasattr(v, '__iter__')


def bytes_(s, encoding='latin-1', errors='strict'):
    """ This function is a copy of ``pyramid.compat.bytes_``

    If ``s`` is an instance of ``text_type``, return
    ``s.encode(encoding, errors)``, otherwise return ``s``"""
    if isinstance(s, str):  # pragma: no cover
        return s.encode(encoding, errors)
    return s


class Sentinel:
    """  This class is a copy of ``pyramid.util.Sentinel``.
    """
    def __init__(self, repr):
        self.repr = repr

    def __repr__(self):
        return self.repr

# These objects are copies of ``pyramid.util.FIRST``, ``pyramid.util.LAST``.
FIRST = Sentinel('FIRST')
LAST = Sentinel('LAST')


class TopologicalSorter:
    """  This class is a copy of ``pyramid.util.TopologicalSorter``.

    A utility class which can be used to perform topological sorts against
    tuple-like data.
    """
    def __init__(self,
                 default_before=LAST,
                 default_after=None,
                 first=FIRST,
                 last=LAST):
        self.names = []
        self.req_before = set()
        self.req_after = set()
        self.name2before = {}
        self.name2after = {}
        self.name2val = {}
        self.order = []
        self.default_before = default_before
        self.default_after = default_after
        self.first = first
        self.last = last

    def remove(self, name):
        """ Remove a node from the sort input """
        self.names.remove(name)
        del self.name2val[name]
        after = self.name2after.pop(name, [])
        if after:
            self.req_after.remove(name)
            for u in after:
                self.order.remove((u, name))
        before = self.name2before.pop(name, [])
        if before:
            self.req_before.remove(name)
            for u in before:
                self.order.remove((name, u))

    def add(self, name, val, after=None, before=None):
        """ Add a node to the sort input.  The ``name`` should be a string or
        any other hashable object, the ``val`` should be the sortable (doesn't
        need to be hashable).  ``after`` and ``before`` represents the name of
        one of the other sortables (or a sequence of such named) or one of the
        special sentinel values :attr:`solo.configurator.util.FIRST`` or
        :attr:`solo.configurator.util.LAST` representing the first or last positions
        respectively.  ``FIRST`` and ``LAST`` can also be part of a sequence
        passed as ``before`` or ``after``.  A sortable should not be added
        after LAST or before FIRST.  An example::

           sorter = TopologicalSorter()
           sorter.add('a', {'a':1}, before=LAST, after='b')
           sorter.add('b', {'b':2}, before=LAST, after='c')
           sorter.add('c', {'c':3})

           sorter.sorted() # will be {'c':3}, {'b':2}, {'a':1}

        """
        if name in self.names:
            self.remove(name)
        self.names.append(name)
        self.name2val[name] = val
        if after is None and before is None:
            before = self.default_before
            after = self.default_after
        if after is not None:
            if not is_nonstr_iter(after):
                after = (after,)
            self.name2after[name] = after
            self.order += [(u, name) for u in after]
            self.req_after.add(name)
        if before is not None:
            if not is_nonstr_iter(before):
                before = (before,)
            self.name2before[name] = before
            self.order += [(name, o) for o in before]
            self.req_before.add(name)

    def sorted(self):
        """ Returns the sort input values in topologically sorted order"""
        order = [(self.first, self.last)]
        roots = []
        graph = {}
        names = [self.first, self.last]
        names.extend(self.names)

        for a, b in self.order:
            order.append((a, b))

        def add_node(node):
            if node not in graph:
                roots.append(node)
                graph[node] = [0]  # 0 = number of arcs coming into this node

        def add_arc(fromnode, tonode):
            graph[fromnode].append(tonode)
            graph[tonode][0] += 1
            if tonode in roots:
                roots.remove(tonode)

        for name in names:
            add_node(name)

        has_before, has_after = set(), set()
        for a, b in order:
            if a in names and b in names:  # deal with missing dependencies
                add_arc(a, b)
                has_before.add(a)
                has_after.add(b)

        if not self.req_before.issubset(has_before):
            raise ConfigurationError(
                'Unsatisfied before dependencies: %s'
                % (', '.join(sorted(self.req_before - has_before)))
            )
        if not self.req_after.issubset(has_after):
            raise ConfigurationError(
                'Unsatisfied after dependencies: %s'
                % (', '.join(sorted(self.req_after - has_after)))
            )

        sorted_names = []

        while roots:
            root = roots.pop(0)
            sorted_names.append(root)
            children = graph[root][1:]
            for child in children:
                arcs = graph[child][0]
                arcs -= 1
                graph[child][0] = arcs
                if arcs == 0:
                    roots.insert(0, child)
            del graph[root]

        if graph:
            # loop in input
            cycledeps = {}
            for k, v in graph.items():
                cycledeps[k] = v[1:]
            raise CyclicDependencyError(cycledeps)

        result = []

        for name in sorted_names:
            if name in self.names:
                result.append((name, self.name2val[name]))

        return result


def viewdefaults(wrapped):
    """ This function is a copy of ``pyramid.util.viewdefaults``.

    Decorator for add_view-like methods which takes into account
    __view_defaults__ attached to view it is passed. Not a documented API but
    used by some external systems.
    """
    def wrapper(self, *arg, **kw):
        defaults = {}
        if arg:
            view = arg[0]
        else:
            view = kw.get('view')
        view = maybe_dotted(view)
        if inspect.isclass(view):
            defaults = getattr(view, '__view_defaults__', {}).copy()

        # We do not use action_method
        # Uncomment this when it's necessary
        # ----------------------------------
        # if not '_backframes' in kw:
        #     kw['_backframes'] = 1 # for action_method

        defaults.update(kw)
        return wrapped(self, *arg, **defaults)
    return functools.wraps(wrapped)(wrapper)


def _pkg_resources_style(value):
    """ This method is taken from Pyramid Web Framework.

    package.module:attr style
    """
    # Calling EntryPoint.load with an argument is deprecated.
    # See https://pythonhosted.org/setuptools/history.html#id8
    ep = pkg_resources.EntryPoint.parse('x={}'.format(value))
    if hasattr(ep, 'resolve'):
        # setuptools>=10.2
        return ep.resolve()  # pragma: NO COVER
    else:
        return ep.load(False)  # pragma: NO COVER


def maybe_dotted(dotted):
    if not isinstance(dotted, str):
        return dotted
    return _pkg_resources_style(dotted)



""" This module is initially taken from Solo

https://github.com/avanov/solo/blob/bf44c527dbe48256d2bd3da463eceeb78d05a38d/solo/configurator/__init__.py
"""
from typing import Optional
from types import ModuleType
import logging
import inspect

import venusian

from .. import registry
from ..util import maybe_dotted
from ..exceptions import ConfigurationError
from ..path import caller_package

from .routes import RoutesConfigurator
from .views import ViewsConfigurator
from .renderers import RenderersConfigurator
from .sums import SumTypesConfigurator


log = logging.getLogger(__name__)


class Configurator:
    """ Every Configurator has its own registry.
    The global app registry this particular configurator produces, is a product of all other configurators combined.
    """
    venusian = venusian
    inspect = inspect

    def __init__(self,
                 routes_configurator=RoutesConfigurator,
                 views_configurator=ViewsConfigurator,
                 renderers_configurator=RenderersConfigurator,
                 sum_types_configurator=SumTypesConfigurator
                 ) -> None:
        self._registry = registry._RegistryBuilder()
        self.registry = None
        self.routes = routes_configurator()
        self.views = views_configurator()
        self.renderers = renderers_configurator()
        self.sums = sum_types_configurator()

    def include(self,
                callable,
                entry_point: 'str' = 'includeme',
                route_prefix: Optional[str] = None,
                namespace: Optional[str] = None) -> None:
        """
        :param callable: package to be configured
        :param route_prefix:
        :return:
        """
        configuration_section: ModuleType = maybe_dotted(callable)
        if namespace is None:
            namespace = configuration_section.__package__
        old_namespace = self.routes.change_namespace(namespace)

        module = self.inspect.getmodule(configuration_section)
        if module is configuration_section:
            try:
                configuration_section = getattr(module, entry_point)
                log.debug(f'Including {callable}:{entry_point}')
            except AttributeError:
                raise ConfigurationError(
                    f"""Package "{module.__name__}" has no entry point called '{entry_point}'. """
                    "Make sure you have defined it."
                )

        sourcefile = self.inspect.getsourcefile(configuration_section)

        if sourcefile is None:
            raise ConfigurationError(
                f'No source file for module {module.__name__} (.py file must exist, '
                f'refusing to use orphan .pyc or .pyo file).'
            )

        if route_prefix is None:
            route_prefix = ''
        route_prefix = f"{self.routes.route_prefix.rstrip('/')}/{route_prefix.lstrip('/')}"
        old_route_prefix = self.routes.change_route_prefix(route_prefix)

        self._log_caption(f'Configuring {callable}:{entry_point}')
        configuration_section(self)
        self.routes.change_namespace(old_namespace)
        self.routes.change_route_prefix(old_route_prefix)

    def scan(self, package=None, categories=None, onerror=None, ignore=None, namespace=None):
        if package is None:
            package = caller_package()

        package = maybe_dotted(package)

        if namespace is None:
            namespace = package.__name__

        self._log_caption(f'Scanning {package}')
        scanner = Scanner(configurator=self)
        previous_namespace = scanner.configurator.routes.change_namespace(namespace)

        scanner.scan(package, categories=categories, onerror=onerror, ignore=ignore)

        self._log_caption('Consistency check')
        self.routes.check_routes_consistency(namespace)
        self.sums.check_sum_types_consistency(namespace)

        self.routes.change_namespace(previous_namespace)
        self._log_caption(f'End scanning {package}')

    def freeze(self) -> registry.AppRegistry:
        self.registry = registry.AppRegistry(*self._registry.as_dict())
        return self.registry

    def _log_caption(self, caption) -> None:
        caption = f' {caption} '
        cap_len = len(caption)
        total_len = 100
        sep_len = int((total_len - cap_len) / 2)
        sep = "-" * sep_len
        log.debug(f'{sep}{caption}{sep}')


class Scanner(venusian.Scanner):
    configurator: Configurator



import os
from typing import List

from pathlib import Path
from setuptools import setup
from setuptools import find_packages


here = Path(os.path.abspath(os.path.dirname(__file__)))


def get_dependencies(path: Path) -> List[str]:
    with path.open() as f:
        rows = f.read().strip().split('\n')
        requires = []
        for row in rows:
            row = row.strip()
            if row and not (row.startswith('#') or row.startswith('http')):
                requires.append(row)
        return requires


with here.joinpath('README.rst').open() as f:
    README = f.read()


install_requires = get_dependencies(here / 'requirements' / 'default.txt')
install_requires.extend(get_dependencies(here / 'requirements' / 'django.txt'))


# Setup
# ----------------------------

setup(name='frameapp',
      version='0.0.3',
      description='Frameapp',
      long_description=README,
      classifiers=[
          'Development Status :: 1 - Planning',
          'Intended Audience :: Developers',
          'License :: OSI Approved',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Operating System :: POSIX',
          'Topic :: Internet :: WWW/HTTP',
      ],
      author='Maxim Avanov',
      author_email='maxim.avanov@gmail.com',
      url='https://github.com/avanov/frameapp',
      keywords='web',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='tests',
      tests_require=get_dependencies(here / 'requirements' / 'testing.txt'),
      install_requires=install_requires,
      entry_points={
      }
    )



""" This module contains exception classes from Pyramid Web Framework.
"""


class ConfigurationError(Exception):
    """ Raised when inappropriate input values are supplied to an API
    method of a :term:`Configurator`
    """


class CyclicDependencyError(Exception):
    """ This class is a copy of ``pyramid.exceptions.CyclicDependencyError``.
    The exception raised when the Pyramid topological sorter detects a
    cyclic dependency.
    """
    def __init__(self, cycles) -> None:
        self.cycles = cycles

    def __str__(self) -> str:
        L = []
        cycles = self.cycles
        for cycle in cycles:
            dependent = cycle
            dependees = cycles[cycle]
            L.append('%r sorts before %r' % (dependent, dependees))
        msg = 'Implicit ordering cycle:' + '; '.join(L)
        return msg



import logging

from django.http import HttpResponse
from django.conf import settings
from pkg_resources import parse_version


log = logging.getLogger(__name__)


class FrameappMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        """ https://docs.djangoproject.com/en/1.11/topics/http/middleware/#process-view
        """
        if 'version' in view_kwargs:
            version = parse_version(view_kwargs['version'])
            if version < settings.FRAMEAPP['MIN_API_VERSION']:
                return HttpResponse(status=410)
        else:
            log.debug(f"API Version is not specified. Defaulting to {settings.FRAMEAPP['MIN_API_VERSION']}")
            version = settings.FRAMEAPP['MIN_API_VERSION']
        setattr(request, 'API_VERSION', version)



import json
from typing import Dict
import pathlib

from django.conf import settings


def json_schema(from_file: str) -> Dict:
    return json.loads(json_schema_str(from_file))


def json_schema_str(from_file: str) -> str:
    path = pathlib.Path(f'{settings.ROOT_PATH}/mobile_api/schemas/{from_file}')
    with path.open('r', encoding='utf-8') as f:
        return f.read()



""" http://www.sphinx-doc.org/en/stable/extdev/appapi.html

This integration allows you to construct a new sphinx directive that would output API docs for registered endpoints.

In your project, create a snippet that would look like this

.. code-block:: python

    # myapp/sphinx_integration.py
    from typing import Dict, Any
    from sphinx.application import Sphinx

    from frameapp import entrypoint
    from frameapp.ext.sphinx_integration.directives import APIDocDirective

    def setup(app: Sphinx) -> Dict[str, Any]:
        config = entrypoint.setup(
            package='myapp',
            apps=<applications>,
            namespace='myapp',
            ignore=<ignore_patterns>
        )
        app.add_directive('myapp-api-doc', APIDocDirective)
        app.add_config_value('FRAMEAPP_CONFIG', config, rebuild='html')
        return {'version': '1.0', 'parallel_read_safe': False}

Then, go to your Sphinx conf.py and find the `extensions = [ ... ]` list and add your newly created extension into it:

.. code-block:: python

    # docs/conf.py
    extensions = [
        'sphinx.ext.autodoc',
        'sphinx.ext.doctest',
        ...
        'myapp.sphinx_integration',
    ]

Now, you can use the directive in your docs

..  code-block::

    =======================
    Endpoints Documentation
    =======================

    .. myapp-api-doc::
        :prefix: /v{version:[0-9]+\.[0-9]+}


"""
